-----------------------------------------------------------------------------
--
--	SIMPLE_UART.VHDL
--
--
--	Author 	: Rene Bllhoff
--  Created : Feb. 2007
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity simple_uart is
  port (clk 				: in 	std_logic;
        res 				: in 	std_logic;

				-- baudrate selection (bit 15 = 1 -> divider; = 0 -> Divider for fixed base-baud-rate amd main divider)
				baudrate		: in std_logic_vector (15 downto 0);

				-- rs232 signals
        rx_in 			: in 	std_logic;
				tx_out			: out std_logic;

				-- transmitter
	      data_in 		: in std_logic_vector(7 downto 0);
	      send  			: in std_logic;
	      ready 			: out std_logic;
		
				-- receiver
        data_out   	: out std_logic_vector (7 downto 0);
        receive			: out std_logic;
				uart_active : out std_logic
			);				  
end simple_uart;

architecture Behavioral of simple_uart is

---------------------------------------------
--
-- UART-Components
--
---------------------------------------------	

	--
	-- UART-Clock Component
	--

	COMPONENT uart_clock
	PORT(
		clk : IN std_logic;
		res : IN std_logic;
		baudrate : IN std_logic_vector(15 downto 0);          
		baud_clk_o : OUT std_logic;
		baud_clk16_o : OUT std_logic
		);
	END COMPONENT;

	--
	-- UART-TX Component
	--

	COMPONENT uart_tx
	PORT(
		clk : IN std_logic;
		clk_baud : IN std_logic;
		res : IN std_logic;
		data_in : IN std_logic_vector(7 downto 0);
		trg : IN std_logic;          
		rdy : OUT std_logic;
		tx_out : OUT std_logic
		);
	END COMPONENT;

	--
	-- UART-RX Component
	--

	COMPONENT uart_rx
	PORT(
		clk : IN std_logic;
		res : IN std_logic;
		rx_in : IN std_logic;
		baud_clk_16 : IN std_logic;          
		busy : OUT std_logic;
		out_buffer : OUT std_logic_vector(7 downto 0);
		trg : OUT std_logic
		);
	END COMPONENT;



	-- Baud Clocks

	signal baud_clk 		: std_logic;
	signal baud_clk_16 	: std_logic;

	-- Ready detection (Edge)
	signal ready_tx			: std_logic;
	signal ready_z1			: std_logic;

	-- Output Buffer for Ready 
	signal ready_o 			: std_logic;

	-- Buffer for UART Active
	signal sending			: std_logic;
	signal receiving		: std_logic;


begin

	--
	-- Generating Baud Clocks
	--
	

	Inst_uart_clock: uart_clock PORT MAP(
		clk  					=> clk,
		res 					=> res,
		baudrate 			=> baudrate,
		baud_clk_o 		=> baud_clk,
		baud_clk16_o 	=> baud_clk_16 
	);

	--
	-- Transmitter
	--
													  
	Inst_uart_tx: uart_tx PORT MAP(
		clk   				=> clk,
		res 					=> res,

		clk_baud 			=> baud_clk,
		tx_out 				=> tx_out,

		trg 					=> send,
		data_in 			=> data_in,
		rdy 					=> ready_tx
	);

	--
	-- Receiver
	--
													  
	Inst_uart_rx: uart_rx PORT MAP(
		clk 					=> clk,
		res 					=> res,

		baud_clk_16 	=> baud_clk_16,
		rx_in 				=> rx_in,

		busy 					=> receiving,
		out_buffer 		=> data_out,
		trg 					=> receive
	);

	-- edge detection for ready

	process (clk,res)
	begin
	  if res = '1' then
			ready_z1 	<= '0';
			ready_o 	<= '0';			
		elsif clk'event and clk = '1' then
			ready_z1 	<= ready_tx;
			ready_o  	<= not ready_z1 and ready_tx;
			sending	 	<= not ready_o;
		end if;
	end process;


	-- output uart active and ready
	uart_active <= sending or receiving;
	ready				<= ready_o;

end Behavioral;
