-----------------------------------------------------------------------------
--
--	UART_TX.VHDL
--
--  Component for transmitting a byte via RS232
--
--	Author 	: Rene Bllhoff
--  Created : Feb. 2007
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity uart_tx is
    Port ( clk : in std_logic;
           res : in std_logic;
					 clk_baud : in std_logic;
           data_in : in std_logic_vector(7 downto 0);
           trg : in std_logic;
           rdy : out std_logic;
           tx_out : out std_logic);
end uart_tx;						

architecture Behavioral of uart_tx is

	-- Registers

	-- Buffer, State 

	signal tx_buffer 		: std_logic_vector (7 downto 0);
	signal state 				: std_logic_vector (3 downto 0);

	-- trigger detection

	signal clk_baud_z1	: std_logic;
	signal trg_z1     	: std_logic;

begin

	process (clk, res, data_in, trg)
	begin
		if res = '1' then

			state 			<= (others => '0');
			tx_buffer 	<= (others => '0');
			tx_out 			<= '1';
			clk_baud_z1 <= '0';
			trg_z1			<= '0';
			rdy					<= '1';

		elsif clk'event and clk = '1' then

			-- for edge detection

			clk_baud_z1 <= clk_baud;
			trg_z1			<= trg;

			-- idle state

			if state = X"0" then
				if trg = '1' and trg_z1 = '0' then	 -- wait for trigger
					state 		<= X"1";
					tx_buffer <= data_in;
					rdy				<= '0';
				else
					rdy				<= '1';
				end if;
				tx_out 		<= '1';
			end if;

			-- trsansmitting states

			if clk_baud_z1 = '0' and clk_baud = '1' then
				case state is
					when X"1" 	=> tx_out <= '0';  						state <= X"2";	rdy				<= '0';	 -- start bit
					when X"2" 	=> tx_out <= tx_buffer (0);  	state <= X"3";	rdy				<= '0';	 -- bit 0
					when X"3" 	=> tx_out <= tx_buffer (1);  	state <= X"4";	rdy				<= '0';
					when X"4" 	=> tx_out <= tx_buffer (2);  	state <= X"5";	rdy				<= '0';
					when X"5" 	=> tx_out <= tx_buffer (3);  	state <= X"6";	rdy				<= '0';
					when X"6" 	=> tx_out <= tx_buffer (4);  	state <= X"7";	rdy				<= '0';
					when X"7" 	=> tx_out <= tx_buffer (5);  	state <= X"8";	rdy				<= '0';
					when X"8" 	=> tx_out <= tx_buffer (6);  	state <= X"9";	rdy				<= '0';
					when X"9" 	=> tx_out <= tx_buffer (7);  	state <= X"a";	rdy				<= '0';  -- bit 7
					when X"a" 	=> tx_out <= '1';            	state <= X"0";	rdy				<= '0';  -- stop bit
					when others => NULL;
				end case;	
			end if;

		end if;
	end process;

end Behavioral;
