-----------------------------------------------------------------------------
--
--  SDSP_CORE_MC.VHDL
--
--	Implements the SDSP-Core. It Holds a MAC-unit, an Address-Generator
--  for Audio-Data and Coefficients, an Audio- (AMEM) and Coefficient-(CMEM) 
--  BlockRAM Instance, a Micro-Code-BlockRAM Instance, several 32-Bit Muxers, 
--  a Statemachine and a hidden-Write Coefficient Mechanism (to reduce Clicks 
--  & Pops on Coefficient Updates)
--
--	Author 	: Rene Bllhoff
--  Created : around Apr. 2006
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity sdsp_core_mc is
    Port ( clk 				: in 	std_logic;
           res 				: in 	std_logic;

					 -- input and output of the SDSP-Core
           input 			: in 	signed(31 downto 0);
           output 		: out signed(31 downto 0);

					 -- Opcode, Parameters and Control-Signals
					 opcode   	: in  std_logic_vector (3 downto 0);	
					 am_adr			: in  std_logic_vector (8 downto 0);
					 cm_adr			: in  std_logic_vector (8 downto 0);
           enable 		: in 	std_logic;
           ready 			: out std_logic;

					 -- Memory Interface (Common Signals for Address, Data, Write Enable)
					 cmi_adr		: in 	std_logic_vector (10 downto 0);
					 cmi_wdata	: in  std_logic_vector ( 7 downto 0);
					 cmi_wen    : in  std_logic;

					 -- Block-RAM Interface for CMEM-BlockRAM
					 cmi_rdata	: out std_logic_vector ( 7 downto 0);
					 cmi_en			: in  std_logic;

					 -- Block-RAM Interface for MMEM-BlockRAM
					 mmi_en			: in  std_logic;
					 mmi_rdata	: out std_logic_vector ( 7 downto 0));
end sdsp_core_mc;

architecture Behavioral of sdsp_core_mc is

---------------------------------------------
--
-- Audio-Data (AMEM) Memory Component
--
---------------------------------------------	

	COMPONENT sdsp_amem
	PORT(
		clk : IN std_logic;
		b_addr : IN std_logic_vector(8 downto 0);
		b_wdata : IN std_logic_vector(31 downto 0);
		b_wen : IN std_logic;
		b_en : IN std_logic;
		a_addr : IN std_logic_vector(8 downto 0);
		a_wdata : IN std_logic_vector(31 downto 0);
		a_wen : IN std_logic;
		a_en : IN std_logic;          
		b_rdata : OUT std_logic_vector(31 downto 0);
		a_rdata : OUT std_logic_vector(31 downto 0)
		);
	END COMPONENT;


---------------------------------------------
--
-- Coefficient (CMEM) Memory Component
--
---------------------------------------------	

	COMPONENT sdsp_cmem
	PORT(
		clk : IN std_logic;
		b_addr : IN std_logic_vector(8 downto 0);
		a_addr : IN std_logic_vector(8 downto 0);
		a_wdata : IN std_logic_vector(31 downto 0);
		a_en : IN std_logic;          
		b_data : OUT std_logic_vector(31 downto 0)
		);
	END COMPONENT;


---------------------------------------------
--
-- MicroCode (MMEM) Memory Component
--
---------------------------------------------	

	COMPONENT sdsp_mmem
	PORT(
		clk : IN std_logic;
		b_addr : IN std_logic_vector(8 downto 0);
		a_addr : IN std_logic_vector(10 downto 0);
		a_wdata : IN std_logic_vector(7 downto 0);
		a_wen : IN std_logic;
		a_en : IN std_logic;          
		b_data : OUT std_logic_vector(31 downto 0);
		a_rdata : OUT std_logic_vector(7 downto 0)
		);
	END COMPONENT;


---------------------------------------------
--
-- 	Signals
--
---------------------------------------------	

	
	--
	-- Audio-Data and Muxer Controls
	--

  -- Audio-Memory	AMEM (write-side)
	signal sdc_a_wdata					: std_logic_vector (31 downto 0);	-- Write AMEM Audio Data

	-- Audio-Memory AMEM (read-side named BMEM)
	signal sdc_b_rdata					: std_logic_vector (31 downto 0);	-- Read BMEM (AMEM) Audio-Data
	
	-- Coefficient-Memory CMEM
	signal sdc_c_rdata					: std_logic_vector (31 downto 0); -- Read CMEM Coefficient Data

	-- Selected Output
  signal sdc_toutput    			: signed (31 downto 0);	 

	-- Data Muxer Controls
	signal sdc_a_data_src 			: std_logic_vector (1 downto 0);	-- Audio-Input at MAC and AMEM
	signal sdc_output_src 			: std_logic_vector (1 downto 0);	-- Output selected from MAC-Output or BMEM
	signal sdc_adr_routing   		: std_logic_vector (1 downto 0); 	-- Address-Routing, BMEM = CMEM (for Load,Store,Move) or seperate AMEM and CMEM (for MAC)
	signal sdc_coeff_routing 		: std_logic_vector (1 downto 0);	-- Coeff-Input at MAC (CMEM, 1.0, 0.0, -1.0)

	-- AMEM Enable 
	signal sdc_a_en    					: std_logic;


	--
	-- Address Generator
	--

	-- Address-Generator Offset Control
	signal sdc_val_a_offs  			: std_logic_vector (3 downto 0);	-- Offset AMEM (from MicroCode)
	signal sdc_val_c_offs	 			: std_logic_vector (3 downto 0);	-- Offset CMEM (from MicroCode)
	signal sdc_b_offset_mode 		: std_logic_vector (1 downto 0);	-- Offset BMEM Select (0,+1,-1)	(-> BMEM = AMEM + Offset)

	-- Address-Generator Set & Increment Controls
	signal sdc_ena_incr_a 		 	: std_logic;											-- Increment AMEM Offset
	signal sdc_ena_incr_c 		 	: std_logic;											-- Increment CMEM Offset
	signal sdc_ena_set_a_offs 	: std_logic;											-- Set sdc_val_a_offs from MicroCode
	signal sdc_ena_set_c_offs 	: std_logic;											-- Set sdc_val_c_offs from MicroCode
																		
	-- Address-Generator Signals (from entity, taken over at new opcode execution)
	signal s_am_adr  						: std_logic_vector (8 downto 0);
	signal s_cm_adr	 						: std_logic_vector (8 downto 0);
	
	-- Address-Generator Signals (counters and static adders
	signal sdc_a_offs						: std_logic_vector (8 downto 0);
	signal sdc_b_offs						: std_logic_vector (8 downto 0);
	signal sdc_c_offs						: std_logic_vector (8 downto 0);

	-- Address-Generator Signals (Results)
	signal sdc_a_addr						: std_logic_vector (8 downto 0);
	signal sdc_b_addr						: std_logic_vector (8 downto 0);
	--signal sdc_b_addr_tmp				: std_logic_vector (8 downto 0);
	signal sdc_c_addr						: std_logic_vector (8 downto 0);


	--
	-- MAC (Multiply and Accumulate) Unit
	--

	-- MAC Ctrl
	signal sdc_ena_macres				: std_logic;	-- Reset MAC-Results
	signal sdc_ena_mac					: std_logic;	-- Enable MAC-Operation (multiply)
	signal sdc_ena_mac_add			: std_logic;	-- Enable MAC-Operation (add)
	signal sdc_ena_coeff_reg    : std_logic;	-- Enable Coeff-Reg

	-- mac data
	signal sdc_mac_coeff 				: signed (31 downto 0);  -- MAC-Input B Coefficient (MAC A Input is sdc_a_wdata)
	signal sdc_mac_val    			: signed (63 downto 0);	 -- MAC-Accu
	signal sdc_mac_val_tmp			: signed (63 downto 0);	 -- MAC-Accu
  signal sdc_mac_val_32 			: signed (31 downto 0);	 -- Clipped MAC-Output (64 bit -> 32 bit)
  signal sdc_mac_out    			: signed (31 downto 0);	 -- Clipped MAC-Output (32 bit -> 24 bit)


	--
	-- SDSP Opcode MicroCode Execution Signals
	--

	-- Main Statemachine
	signal sdc_microcode_state	: std_logic_vector (7 downto 0);

	-- MicroCode BLockRAM Signals (Address and Read Data)
	signal sdc_m_addr						: std_logic_vector ( 8 downto 0);	
	signal sdc_m_rdata 					: std_logic_vector (31 downto 0);

	-- Execution Ready-Signal
	signal sdc_ready_t					: std_logic;

	-- Enable Opcode Signals
	signal sdc_enable_mc				: std_logic;


	--
	-- CMEM Hidden Write Mechanism
	--

	-- 8 x 32 Bit Registers
	signal cm_hidden_data 			: std_logic_vector (31 downto 0);
	signal cm_hidden_data1			: std_logic_vector (31 downto 0);
	signal cm_hidden_data2			: std_logic_vector (31 downto 0);
	signal cm_hidden_data3			: std_logic_vector (31 downto 0);
	signal cm_hidden_data4			: std_logic_vector (31 downto 0);
	signal cm_hidden_data5			: std_logic_vector (31 downto 0);
	signal cm_hidden_data6			: std_logic_vector (31 downto 0);
	signal cm_hidden_data7			: std_logic_vector (31 downto 0);
	signal cm_hidden_data8			: std_logic_vector (31 downto 0);

	-- Address to Write to CMEM
	signal cmih_addr						: std_logic_vector ( 8 downto 0); -- Resulting Address (cmih_addr = cm_hidden_addr + cm_hidden_offs)
	-- CMEM-BlockRAM Enable Signal
	signal cmih_en   						: std_logic;

	-- Hidden CMEM Write State Machine Signal and State Machine Trigger Signal
	signal cm_hidden_state 		  : std_logic_vector ( 3 downto 0);
	signal cm_hidden_wr					: std_logic;											-- Hidden Write Trigger
	signal cm_hidden_wr_z1			: std_logic;											-- Hidden Write Trigger (Delayed for Edge Detect)
	
	-- Hidden CMEM Write Registers, Address Generator and CMEM-Write Edge Detect
	signal cm_hidden_addr				: std_logic_vector ( 8 downto 0);	 -- Base Write Address
	signal cm_hidden_offs				: std_logic_vector ( 2 downto 0);	 -- Offset Counter
	signal cm_hidden_len 				: std_logic_vector ( 2 downto 0);	 -- Length-Register


begin

---------------------------------------------
--
-- 	Instanciation of components
--
---------------------------------------------	

	--
	-- Instance of Audio-Data-BlockRAM AMEM
	--

	Inst_sdsp_amem: sdsp_amem PORT MAP(
		clk 		=> clk,
		b_addr 	=> sdc_b_addr,					 -- B-Side is ALWAYS Read and ALWAYS active)
		b_wdata => (others => '0'),
		b_rdata => sdc_b_rdata,
		b_wen 	=> '0',
		b_en 		=> '1',
		a_addr 	=> sdc_a_addr,					 -- B-Side is ALWAYS Write and active when sdc_a_en = '1')
		a_wdata => sdc_a_wdata,
		a_rdata => open,
		a_wen 	=> '1',
		a_en 		=> sdc_a_en
	);

	--
	-- Instance of Coefficient-BlockRAM CMEM
	--

	Inst_sdsp_cmem: sdsp_cmem PORT MAP(
		clk 		=> clk,
		b_addr 	=> sdc_c_addr,
		b_data 	=> sdc_c_rdata,
		a_addr 	=> cmih_addr,
		a_wdata => cm_hidden_data,
		a_en 		=> cmih_en
	);

	--
	-- Instance of MicroCode-BlockRAM MMEM
	--

	Inst_sdsp_mmem: sdsp_mmem PORT MAP(
		clk 		=> clk,
		b_addr 	=> sdc_m_addr,
		b_data 	=> sdc_m_rdata,
		a_addr 	=> cmi_adr, 
		a_wdata => cmi_wdata,
		a_rdata => mmi_rdata,
		a_wen 	=> cmi_wen,
		a_en 		=> mmi_en
	);



---------------------------------------------
--
-- Processes and logic
--
---------------------------------------------	


	---------------------------------------------
	--
	-- Hidden CMEM-Write
	--
	---------------------------------------------	


	--
	-- Interfacing Hidden Write Registers to CMEM-BlockRAM (Registers are Write Only and Accessed via Memory Mapped CMEM-Writes)
	--

	process (clk, res, cmi_wdata, cmi_adr, cmi_en)
	begin
	  if res = '1' then
		  cm_hidden_data1 <= (others => '0');
		  cm_hidden_data2 <= (others => '0');
		  cm_hidden_data3 <= (others => '0');
		  cm_hidden_data4 <= (others => '0');
		  cm_hidden_data5 <= (others => '0');
		  cm_hidden_data6 <= (others => '0');
		  cm_hidden_data7 <= (others => '0');
		  cm_hidden_data8 <= (others => '0');

			cm_hidden_addr  <= (others => '0');
			cm_hidden_len   <= (others => '0');
			cm_hidden_wr 		<= '0';


		elsif clk'event and clk = '1' then
		  -- if write operation ...
			if cmi_en = '1' then
			  case cmi_adr (7 downto 0) is
				  -- write data (8 Bit) to hidden write registers (32 bit)
					when X"00" => cm_hidden_data1 ( 7 downto  0) <= cmi_wdata;
					when X"01" => cm_hidden_data1 (15 downto  8) <= cmi_wdata;
					when X"02" => cm_hidden_data1 (23 downto 16) <= cmi_wdata;
					when X"03" => cm_hidden_data1 (31 downto 24) <= cmi_wdata;

					when X"04" => cm_hidden_data2 ( 7 downto  0) <= cmi_wdata;
					when X"05" => cm_hidden_data2 (15 downto  8) <= cmi_wdata;
					when X"06" => cm_hidden_data2 (23 downto 16) <= cmi_wdata;
					when X"07" => cm_hidden_data2 (31 downto 24) <= cmi_wdata;

					when X"08" => cm_hidden_data3 ( 7 downto  0) <= cmi_wdata;
					when X"09" => cm_hidden_data3 (15 downto  8) <= cmi_wdata;
					when X"0a" => cm_hidden_data3 (23 downto 16) <= cmi_wdata;
					when X"0b" => cm_hidden_data3 (31 downto 24) <= cmi_wdata;

					when X"0c" => cm_hidden_data4 ( 7 downto  0) <= cmi_wdata;
					when X"0d" => cm_hidden_data4 (15 downto  8) <= cmi_wdata;
					when X"0e" => cm_hidden_data4 (23 downto 16) <= cmi_wdata;
					when X"0f" => cm_hidden_data4 (31 downto 24) <= cmi_wdata;

					when X"10" => cm_hidden_data5 ( 7 downto  0) <= cmi_wdata;
					when X"11" => cm_hidden_data5 (15 downto  8) <= cmi_wdata;
					when X"12" => cm_hidden_data5 (23 downto 16) <= cmi_wdata;
					when X"13" => cm_hidden_data5 (31 downto 24) <= cmi_wdata;

					when X"14" => cm_hidden_data6 ( 7 downto  0) <= cmi_wdata;
					when X"15" => cm_hidden_data6 (15 downto  8) <= cmi_wdata;
					when X"16" => cm_hidden_data6 (23 downto 16) <= cmi_wdata;
					when X"17" => cm_hidden_data6 (31 downto 24) <= cmi_wdata;

					when X"18" => cm_hidden_data7 ( 7 downto  0) <= cmi_wdata;
					when X"19" => cm_hidden_data7 (15 downto  8) <= cmi_wdata;
					when X"1a" => cm_hidden_data7 (23 downto 16) <= cmi_wdata;
					when X"1b" => cm_hidden_data7 (31 downto 24) <= cmi_wdata;

					when X"1c" => cm_hidden_data8 ( 7 downto  0) <= cmi_wdata;
					when X"1d" => cm_hidden_data8 (15 downto  8) <= cmi_wdata;
					when X"1e" => cm_hidden_data8 (23 downto 16) <= cmi_wdata;
					when X"1f" => cm_hidden_data8 (31 downto 24) <= cmi_wdata;

					-- adress & length and trigger registers
					when X"20" => cm_hidden_addr 	(	7 downto 	0) <= cmi_wdata;
					when X"21" => cm_hidden_len  <= cmi_wdata (2 downto 0);
					when X"22" => cm_hidden_wr   <= '1'; -- trick : simplify write to reg 0x22 and 0x23 to trigger operation
					when X"23" => cm_hidden_wr   <= '0'; -- cm_hidden register can only be reset to write something to these registers
																							 -- in addition with autoincrement by DEVICE_6xBRAM_IOMEM can trigger be taken back easily

					when others => NULL;
				end case;
			end if;
		end if;
		
	end process;

	--
	-- Muxer for Output-Registers to CMEM-BlockRAM Write-Side
	--

	with cm_hidden_offs select
		cm_hidden_data <= cm_hidden_data1 when "000",
											cm_hidden_data2 when "001",
											cm_hidden_data3 when "010",
											cm_hidden_data4 when "011",
											cm_hidden_data5 when "100",
											cm_hidden_data6 when "101",
											cm_hidden_data7 when "110",
											cm_hidden_data8 when "111",
											(others => '0') when others;

	--
	-- Address Generation for Hidden CMEM Write (static Adder)
	--

	cmih_addr <= cm_hidden_addr + cm_hidden_offs;

	--
	-- Statemachine for Controlling Hidden CMEM Write
	--

	process (clk, res, cm_hidden_state, cm_hidden_wr_z1, cm_hidden_wr, cm_hidden_offs, cm_hidden_len)
	begin
		if res = '1' then
			cm_hidden_wr_z1 <= '0';
			cm_hidden_offs 	<= (others => '0');
			cmih_en 				<= '0';
			cm_hidden_state	<= (others => '0');

		elsif clk'event and clk = '1' then

			-- Edge detect for Hidden CMEM Write Trigger Signal
			cm_hidden_wr_z1 <= cm_hidden_wr;

			-- Hidden CMEM Write Statemachine
			-- State 0 : Wait for Edge detect of Trigger Signal	and Reset Counter
			-- State 1 : Wait Cycle
			-- State 2 : Enable CMEM Write
			-- State 3 : Disable CMEM Write
			-- State 4 : Increment Offset Register
			-- State 5 : Wait Cycle
			-- State 6 : Check if Values to Copy (-> State 1) or End of Hidden Write (-> State 0)
			case cm_hidden_state is
				-- State 0 : Wait for Edge detect of Trigger Signal	and Reset Counter
				when X"0" => 	if cm_hidden_wr_z1 = '0' and cm_hidden_wr = '1' then
												cm_hidden_state <= X"1";
												cm_hidden_offs 	<= (others => '0');
											end if;
				-- State 1 : Wait Cycle
				when X"1" => 	cmih_en <= '0';
											cm_hidden_state <= X"2";
				-- State 2 : Enable CMEM Write
				when X"2" => 	cmih_en <= '1';
											cm_hidden_state <= X"3";
				-- State 3 : Disable CMEM Write
				when X"3" => 	cmih_en <= '0';
											cm_hidden_state <= X"4";
				-- State 4 : Increment Offset Register
				when X"4" => 	cm_hidden_offs <= cm_hidden_offs + '1';
											cm_hidden_state <= X"5";
				-- State 5 : Wait Cycle
				when X"5" => 	cm_hidden_state <= X"6";
				-- State 6 : Check if Values to Copy (-> State 1) or End of Hidden Write (-> State 0)
				when X"6" => 	if cm_hidden_offs < cm_hidden_len then
											  cm_hidden_state <= X"1";
											else
												cm_hidden_state <= X"0";
											end if;
				when others => NULL;
			end case;

		end if;
	end process;







	---------------------------------------------
	--
	-- Address Generator
	--
	---------------------------------------------	

	--
	-- Offset Counter for AMEM and CMEM and Offset Select for BMEM
	--

	process (clk, res)
	begin
	  if res = '1' then
 		  sdc_a_offs <= (others => '0');
 		  sdc_b_offs <= (others => '0');
 		  sdc_c_offs <= (others => '0');

			--sdc_b_addr_tmp <= (others => '0');

		elsif clk'event and clk = '1' then

			-- BMEM Offset Select
			case sdc_b_offset_mode is
				when "00" 	=> sdc_b_offs <= "000000000";
				when "01" 	=> sdc_b_offs <= "000000001";
				when "10"		=> sdc_b_offs <= "111111111";
				when others => NULL;
			end case;

			-- Increment AMEM Offset
			if sdc_ena_incr_a = '1' then
				sdc_a_offs <= sdc_a_offs + '1';
			end if;

			-- Increment CMEM Offset
			if sdc_ena_incr_c = '1' then
				sdc_c_offs <= sdc_c_offs + '1';
			end if;

			-- Set AMEM Offset Counter Value through MicroCode
			if sdc_ena_set_a_offs = '1' then
				sdc_a_offs <= "00000" & sdc_val_a_offs;
			end if;

			-- Set CMEM Offset Counter Value through MicroCode
			if sdc_ena_set_c_offs = '1' then
				sdc_c_offs <= "00000" & sdc_val_c_offs;
			end if;
		end if;
	end process;



	--
	-- Address Generation (3 static Adders and 1 Muxer)
	--

	-- AMEM- and CMEM-Address always Base-Address + Offset
	sdc_a_addr <= s_am_adr 	+ sdc_a_offs;
	sdc_c_addr <= s_cm_adr 	+ sdc_c_offs;
	-- BMEM can be sdc_a_addr + BMEM Offset or s_cm_adr (for Load,Store,Move)

	--sdc_b_addr <= sdc_a_addr + sdc_b_offs when sdc_adr_routing = "00" else s_cm_adr;
	with sdc_adr_routing select
		 sdc_b_addr <= sdc_a_addr + sdc_b_offs when "00",
									 s_cm_adr								 when "01",
									 s_am_adr								 when "10",
									 (others => '0')				 when others;


	---------------------------------------------
	--
	-- 	Data Muxers (Input and Coefficient)
	--
	---------------------------------------------	


	--
	-- Muxer for MAC-Input A (Audio-Data) and AMEM-Write-Data
	--

	process (clk, res, sdc_a_data_src, input, sdc_b_rdata, sdc_mac_out)
	begin
	  if res = '1' then
			sdc_a_wdata <= (others => '0');
		elsif clk'event and clk = '1' then
			case sdc_a_data_src is
				when "00"		=> sdc_a_wdata <= std_logic_vector (input);
				when "01"		=> sdc_a_wdata <= sdc_b_rdata;
				when "10"		=> sdc_a_wdata <= std_logic_vector (sdc_mac_val_32);
				when "11"		=> sdc_a_wdata <= (others => '0');
				when others	=> sdc_a_wdata <= (others => '0');
			end case;
	  end if;
  end process;

	--
	-- Muxer for Output of SDSP Core (MAC-Output or BMEM)
	--

	process (clk, res, sdc_output_src)
	begin
	  if res = '1' then
		  sdc_toutput <= (others => '0');
		elsif clk'event and clk = '1' then
			case sdc_output_src is
				when "00"   => sdc_toutput <= sdc_mac_out;
				when "01"   => sdc_toutput <= signed (sdc_b_rdata);
  			when others => NULL;
			end case;
    end if;
  end process;


	--
	-- Coeff-Muxer (can be CMEM, 1.0, 0.0, -1.0 )
	--

	process (clk, res, sdc_c_rdata, sdc_coeff_routing, sdc_ena_coeff_reg)
	begin
	  if res = '1' then
		  sdc_mac_coeff <= X"00000000";
		elsif clk'event and clk = '1' then
		  if sdc_ena_coeff_reg = '1' then
				case sdc_coeff_routing is
				  when "00"   => sdc_mac_coeff <= signed (sdc_c_rdata);
				  when "01"   => sdc_mac_coeff <= X"01000000";
				  when "10"   => sdc_mac_coeff <= X"00000000";
				  when "11"   => sdc_mac_coeff <= X"FFFFFFFF";
					when others => NULL;
				end case;
			end if;
		end if;
	end process;


	---------------------------------------------
	--
	-- MAC-Unit
	--
	---------------------------------------------	


	--
	-- MAC-Unit itself
	--

	process (clk, res, sdc_ena_macres, sdc_ena_mac, sdc_mac_val)
	begin
		if res = '1' then
			sdc_mac_val   <= (others => '0');
		elsif clk'event and clk = '1' then

			-- Reset Accumulator
			if sdc_ena_macres = '1' then
				sdc_mac_val   	<= (others => '0');
				sdc_mac_val_tmp	<= (others => '0');
			end if;

			-- Enable Operation
			if sdc_ena_mac = '1' then
			  sdc_mac_val_tmp <= ( signed (sdc_a_wdata) * signed (sdc_mac_coeff) );
			end if;
		  -- Enable Acuumulate
			if sdc_ena_mac_add = '1' then
			  sdc_mac_val <= sdc_mac_val + sdc_mac_val_tmp;
			end if;
		end if;
	end process;


	--
	-- Hard-Clipping of 64-Bit Result of MAC-Accumulator
	--
	-- Clipping is done on the bits 55..24 (we have 8 bit headroom at multiply)
	--

	process (clk, res, sdc_mac_val)
	begin
	  if res = '1' then
      sdc_mac_val_32 <= (others => '0');
	  elsif clk'event and clk = '1' then
			-- dependend on sign
			if sdc_mac_val (63) = '1' then
			  -- check only the 9 highest bits of relevant data 
				if std_logic_vector (sdc_mac_val (63 downto 55)) = "111111111" then
				  -- in range
				 	sdc_mac_val_32 <= sdc_mac_val (55 downto 24);
				else
				  -- out of range
				 	sdc_mac_val_32 <= X"80000000";
				end if;
			else
			  -- check only the 9 highest bits of relevant data
				if std_logic_vector (sdc_mac_val (63 downto 55)) = "000000000" then
				  -- in range
				 	sdc_mac_val_32 <= sdc_mac_val (55 downto 24);
				else
				  -- out of range
				 	sdc_mac_val_32 <= X"7fffffff";
				end if;
			end if;
		end if;
	end process;


	--
	-- Hard-Clipping of 32-Bit Result to 24 Bit Result (for I2S-Output is 24 Bit Wide, may add an Enable signal to give mre control over clipping)
	--

	process (clk, res, sdc_mac_val_32)
	begin
	  if res = '1' then
      sdc_mac_out <= (others => '0');
	  elsif clk'event and clk = '1' then
		  -- dependend on sign
			if sdc_mac_val_32 (31) = '1' then
			  -- check only the 9 highest bits of relevant data 
				if std_logic_vector (sdc_mac_val_32 (31 downto 23)) = "111111111" then
					-- in range
				 	sdc_mac_out <= sdc_mac_val_32;
				else
				  -- out of range
				 	sdc_mac_out <= X"ff800000";
				end if;
			else
			  -- check only the 9 highest bits of relevant data
				if std_logic_vector (sdc_mac_val_32 (31 downto 23)) = "000000000" then
					-- in range
				 	sdc_mac_out <= sdc_mac_val_32;
				else
				  -- out of range
				 	sdc_mac_out <= X"007fffff";
				end if;
			end if;
		end if;
	end process;




	---------------------------------------------
	--
	-- Microcode Execution and Main Statemachine
	--
	---------------------------------------------	


	--
	-- "Decode" Data from MicroCode-BlockRAM when sdc_enable_mc = '1'
	--
 
	sdc_ena_incr_a 				<= sdc_m_rdata (27)							when sdc_enable_mc = '1' else '0';
	sdc_ena_coeff_reg			<= sdc_m_rdata (26)							when sdc_enable_mc = '1' else '0';
	sdc_ena_mac_add	 			<= sdc_m_rdata (25)							when sdc_enable_mc = '1' else '0';

	-- old microcode decode

	sdc_ena_set_a_offs		<= sdc_m_rdata (24)							when sdc_enable_mc = '1' else '0';
	sdc_ena_set_c_offs		<= sdc_m_rdata (23)							when sdc_enable_mc = '1' else '0';
	sdc_ena_incr_c				<= sdc_m_rdata (22)							when sdc_enable_mc = '1' else '0';
	sdc_b_offset_mode			<= sdc_m_rdata (21 downto 20)		when sdc_enable_mc = '1' else (others => '0');
	sdc_adr_routing				<= sdc_m_rdata (18 downto 17)  	when sdc_enable_mc = '1' else (others => '0');
  sdc_val_a_offs				<= sdc_m_rdata (16 downto 13)		when sdc_enable_mc = '1' else (others => '0');
	sdc_val_c_offs				<= sdc_m_rdata (12 downto  9)		when sdc_enable_mc = '1' else (others => '0');	
	sdc_a_en							<= sdc_m_rdata (8)							when sdc_enable_mc = '1' else '0';
													 
	sdc_a_data_src				<= sdc_m_rdata ( 7 downto  6)		when sdc_enable_mc = '1' else (others => '0');	
	sdc_coeff_routing			<= sdc_m_rdata ( 5 downto  4)		when sdc_enable_mc = '1' else (others => '0');	
	sdc_ena_macres				<= sdc_m_rdata (3)							when sdc_enable_mc = '1' else '0';	
	sdc_ena_mac						<= sdc_m_rdata (2)							when sdc_enable_mc = '1' else '0';	
	sdc_output_src				<= '0' & sdc_m_rdata (1)				when sdc_enable_mc = '1' else (others => '0');	
													 															 
	sdc_ready_t						<= sdc_m_rdata (0)							when sdc_enable_mc = '1' else '0';
																																  
	ready									<= sdc_ready_t									when sdc_enable_mc = '1' else '0';



	-- MicroCode Execution Statemachine
	process (clk, res, sdc_microcode_state, enable)
	begin
	  if res = '1' then

			sdc_microcode_state	<= (others => '0');
			sdc_m_addr 					<= (others => '0');
			s_am_adr 						<= (others => '0');
			s_cm_adr 						<= (others => '0');
		  output 							<= (others => '0');

		elsif	clk'event and clk = '1' then

			case sdc_microcode_state is
				-- Wait for Enable through High Level Instance
				when X"00"  => 	if enable = '1' then
													-- "Calculate" Start Address of Micro-Code
													sdc_m_addr 					<= opcode & "00000";
													sdc_enable_mc				<= '0';	
													sdc_microcode_state <= X"01";
													-- take over memory pointers for AMEM and CMEM (or BMEM)
													s_am_adr 						<= am_adr;
													s_cm_adr 						<= cm_adr;
												else
												  sdc_m_addr 					<= '0' & X"00"; 
													sdc_microcode_state <= X"00";
													sdc_enable_mc				<= '0';
												end if;

				-- Enable MicroCode Control (MicroCode Data is stable and can Control SDSP)
				when X"01"  =>  sdc_microcode_state 	<= X"02";
												sdc_enable_mc					<= '1';

				-- Run until sdc_ready_t = '1', otherwise disable MicroCode Control and Reset MicroCode Address
				when X"02"  =>  if sdc_ready_t = '1' then
												  sdc_m_addr 					<= (others => '0');
													sdc_microcode_state <= X"00";
													sdc_enable_mc				<= '0';
													output							<= sdc_toutput;
												else
													sdc_m_addr <= sdc_m_addr + '1';
													if sdc_m_addr = '1' & X"FF" then
 													  sdc_microcode_state	<= X"00";
														sdc_enable_mc				<= '0';
													end if;
												end if;

				when others => 	NULL;
			end case;

		end if;

	end process;

end Behavioral;
