-----------------------------------------------------------------------------
--
--  UTIL_CORE.VHDL
--
--	Implements the Utility Core for Generation of Audio-Signals 
--  (currently only Oscillators). It Holds an accumulator,
--  Waveshaper, a Statemachine for controlling all internal sequences
--  as well as an Micro-Code - BlockRAM (UTIL_MMEM), and a Data-BlockRAM 
--  (UTIL_DMEM).
--
--	Author 	: Rene Bllhoff
--  Created : around Oct. 2006
--
--	(c) 2005-2007 by Rene Bllhoff - reneboellhoff (at) gmx (dot) de - 
-----------------------------------------------------------------------------
--
--  This source file may be used and distributed without
--  restriction provided that this copyright statement is not
--  removed from the file and that any derivative work contains
--  the original copyright notice and the associated disclaimer.
--
--  This program is free software; you can redistribute it and/or
--  modify it under the terms of the GNU General Public License
--  as published by the Free Software Foundation; either version 2
--  of the License, or (at your option) any later version.
--										
--  This program is distributed in the hope that it will be useful,
--  but WITHOUT ANY WARRANTY; without even the implied warranty of
--  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU
--  General Public License for more details.
--
--  You should have received a copy of the GNU General Public
--  License along with this source; if not, download it	from
--  http://www.gnu.org/licenses/gpl.html	
--
-----------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;


entity util_core is
    Port ( clk 				: in 	std_logic;
           res 				: in 	std_logic;

					 -- input and output of the Util-Core
           input 			: in 	signed(31 downto 0);
           output 		: out signed(31 downto 0);

					 -- Opcode, Parameters and Control-Signals
           opcode 		: in 	std_logic_vector ( 3 downto 0);
           am_adr 		: in 	std_logic_vector ( 8 downto 0);
           cm_adr 		: in 	std_logic_vector ( 8 downto 0);
           enable 		: in 	std_logic;
           ready 			: out std_logic;

					 -- SRAM-Access IO (Spartan 3 On Board  Hardware)
					 sri_data 	: INOUT std_logic_vector(15 downto 0);      
					 sri_addr 	: OUT std_logic_vector(17 downto 0);
					 sri_cs1 		: OUT std_logic;
					 sri_cs2 		: OUT std_logic;
					 sri_oe 		: OUT std_logic;
					 sri_ub 		: OUT std_logic;
					 sri_lb 		: OUT std_logic;
					 sri_we 		: OUT std_logic;

					 -- Memory Interface (Common Signals for Address, Data, Write Enable)						
					 dmi_adr		: in 	std_logic_vector (10 downto 0);
					 dmi_wdata	: in 	std_logic_vector ( 7 downto 0);
					 dmi_wen		: in 	std_logic;

					 -- Block-RAM Interface for Data Memory DMEM
					 dmi_rdata	: out	std_logic_vector ( 7 downto 0);
					 dmi_en			: in 	std_logic;

					 -- Block-RAM Interface for MicroCore MMEM
					 mmi_rdata	: out	std_logic_vector ( 7 downto 0);
					 mmi_en			: in 	std_logic;

					 -- Block-RAM Interface for Wave Memory WMEM
					 wmi_rdata	: out	std_logic_vector ( 7 downto 0);
					 wmi_en			: in 	std_logic
					 );
end util_core;

architecture Behavioral of util_core is

---------------------------------------------
--
-- Data Memory (DMEM) Component
--
---------------------------------------------	

	COMPONENT util_dmem
	PORT(
		clk 		: IN 	std_logic;
		b_addr 	: IN 	std_logic_vector(8 downto 0);
		b_wdata : IN 	std_logic_vector(31 downto 0);
		b_rdata : OUT std_logic_vector(31 downto 0);
		b_wen 	: IN 	std_logic;
		b_en 		: IN 	std_logic;
		a_addr 	: IN 	std_logic_vector(10 downto 0);
		a_wdata : IN 	std_logic_vector( 7 downto 0);
		a_rdata : OUT std_logic_vector( 7 downto 0);
		a_wen 	: IN 	std_logic;
		a_en 		: IN 	std_logic
		);
	END COMPONENT;

---------------------------------------------
--
-- MicroCode (MMEM) Component
--
---------------------------------------------	

	COMPONENT util_mmem
	PORT(
		clk 		: IN 	std_logic;
		b_addr 	: IN 	std_logic_vector(8 downto 0);
		a_addr 	: IN 	std_logic_vector(10 downto 0);
		a_wdata : IN 	std_logic_vector(7 downto 0);
		a_wen 	: IN 	std_logic;
		a_en 		: IN 	std_logic;          
		b_data 	: OUT std_logic_vector(31 downto 0);
		a_rdata : OUT std_logic_vector(7 downto 0)
		);
	END COMPONENT;



---------------------------------------------
--
-- Wave Memory (WMEM) Component
--
---------------------------------------------	

	COMPONENT util_wmem
	PORT(
		clk : IN std_logic;
		b_addr : IN std_logic_vector(9 downto 0);
		a_addr : IN std_logic_vector(10 downto 0);
		a_wdata : IN std_logic_vector(7 downto 0);
		a_wen : IN std_logic;
		a_en : IN std_logic;          
		b_data : OUT std_logic_vector(15 downto 0);
		a_rdata : OUT std_logic_vector(7 downto 0)
		);
	END COMPONENT;


---------------------------------------------
--
-- SRAM Access Compoment
--
---------------------------------------------	

	COMPONENT sram_access
	PORT(
		clk : IN std_logic;
		res : IN std_logic;
		ri_adr : IN std_logic_vector(17 downto 0);
		ri_din : IN std_logic_vector(15 downto 0);
		ri_rw : IN std_logic;
		ri_trg : IN std_logic;    
		sri_data : INOUT std_logic_vector(15 downto 0);      
		sri_adr : OUT std_logic_vector(17 downto 0);
		sri_cs1 : OUT std_logic;
		sri_cs2 : OUT std_logic;
		sri_oe : OUT std_logic;
		sri_ub : OUT std_logic;
		sri_lb : OUT std_logic;
		sri_we : OUT std_logic;
		ri_dout : OUT std_logic_vector(15 downto 0);
		ri_rdy : OUT std_logic
		);
	END COMPONENT;


---------------------------------------------
--
-- Signals
--
---------------------------------------------	

	--
	-- Muxer
	--

	-- Input Muxer 

	signal in_src							: std_logic_vector (1 downto 0);
	signal in_muxer       		: signed (31 downto 0);


	-- Output Muxer 

	signal out_src						: std_logic_vector (1 downto 0);		 	
	signal out_muxer      		: signed (31 downto 0);

	-- DMEM Write Data Muxer 

	signal dmem_wdata_src			: std_logic_vector (1 downto 0);		 	
	signal dmem_wdata 				: signed (31 downto 0);



	--
	-- Oscillator (Acuu & Shaper)
	--

	-- Accumulator & Ctrls

	signal accu     					: signed (31 downto 0);
	signal ena_accu_load 			: std_logic;
	signal ena_accu      			: std_logic;
	signal accu_load_src			: std_logic_vector (1 downto 0);
	

	-- Wave Whaper, Controls & WREG (for Shape Select)

	signal shape_out  				: signed (31 downto 0);
	signal ena_shaper					: std_logic;
	--signal shape_sel					: std_logic_vector ( 1 downto 0);

	-- Data Register (for Shaper)
	signal wave_reg_en				: std_logic;
	signal wave_reg						: std_logic_vector (31 downto 0);

	-- Wave Memory
	signal wmem_adr						: std_logic_vector (9 downto 0);
	signal wmem_data					: std_logic_vector (15 downto 0);


	-- Noise Register

	signal noise_reg					: std_logic_vector (31 downto 0);
	signal noise_shift_ena		: std_logic;
	signal noise_input				: std_logic;




	--
	-- SRAM Address Generator and Controls
	--

  -- Data Interface from/to SRAM Access Statemachine
	signal sr_din		: std_logic_vector (15 downto 0);		
	signal sr_dout	: std_logic_vector (15 downto 0);		
	
	-- Trigger and control Signals for Statemachine
	signal sr_trg 	: std_logic;
	signal sr_wen 	: std_logic;

	-- Address Generator Signals (Values)

	-- Base Address
  signal sram_base_addr 	: std_logic_vector (17 downto 0);	
	-- Counter
	signal sram_base_count  : std_logic_vector (17 downto 0);
	-- Offset (to counter)
  signal sram_base_offset : std_logic_vector (17 downto 0);
	-- Mask
	signal sram_offs_mask   : std_logic_vector (17 downto 0);
	-- Resulting Adress
	signal sram_eff_addr    : std_logic_vector (17 downto 0);

	-- Address Generator Signals (Control)

	-- Load Address Generator Value (Enable)
	signal load_sram_addr		: std_logic;
	-- Load Address Generator Value (Destination)
	signal load_sram_dest		: std_logic_vector (1 downto 0);
	-- Clear Offset Register
	signal clear_sram_offs	: std_logic;
	-- Increment Counter
	signal inc_sram_addr		: std_logic;



	--
	-- DMEM Address Generator and Controls
	--

	-- DMEM Address Generator

	signal dm_offs            : std_logic_vector ( 3 downto 0);
	signal ena_set_dmem_offs	: std_logic;
	signal ena_dmem_offs_incr	: std_logic;
	signal dmem_adr_offs			: std_logic_vector ( 3 downto 0);


	-- DMEM Interface (core-side)

	signal s_dm_adr        		: std_logic_vector ( 8 downto 0);
	signal dm_adr          		: std_logic_vector ( 8 downto 0);
	signal dm_rdata        		: std_logic_vector (31 downto 0);
	signal dm_wen          		: std_logic;
	signal dm_en           		: std_logic;


																
	--
	-- MicroCode Statemachine
	--

	-- MicroCode Statemachine-Signals
	signal enable_mc					: std_logic;
	signal umc_state 					: std_logic_vector (3 downto 0);
	signal ready_t						: std_logic;

	-- MicroCode BlockRAM
	signal mm_adr	 						: std_logic_vector ( 8 downto 0);
	signal mm_rdata						: std_logic_vector (31 downto 0);


begin

---------------------------------------------
--
-- Instanciation of components
--
---------------------------------------------	

	--
	-- Instance of Data-BlockRAM DMEM
	--

	Inst_util_dmem: util_dmem PORT MAP(
		clk 		=> clk,
		-- Access From Util Core
		b_addr 	=> dm_adr,
		b_wdata => std_logic_vector (dmem_wdata),
		b_rdata => dm_rdata,
		b_wen 	=> dm_wen,
		b_en 		=> dm_en,
		-- Access from Memory Interface
		a_addr 	=> dmi_adr,
		a_wdata => dmi_wdata,
		a_rdata => dmi_rdata,
		a_wen 	=> dmi_wen,
		a_en 		=> dmi_en
	);

	--
	-- Instance of MicroCode-BlockRAM MMEM
	--

	Inst_util_mmem: util_mmem PORT MAP(
		clk 		=> clk,
		-- Access From Util Core
		b_addr 	=> mm_adr,
		b_data 	=> mm_rdata,
		-- Access from Memory Interface
		a_addr 	=> dmi_adr,
		a_wdata => dmi_wdata,
		a_wen 	=> dmi_wen,
		a_rdata => mmi_rdata,
		a_en 		=> mmi_en
	);

	--
	-- Instance of Wave Memory-BlockRAM WMEM
	--

	Inst_util_wmem: util_wmem PORT MAP(
		clk 		=> clk,
		-- Access From Util Core
		b_addr	=> wmem_adr,
		b_data 	=> wmem_data,
		-- Access from Memory Interface
		a_addr 	=> dmi_adr,
		a_wdata => dmi_wdata,
		a_wen 	=> dmi_wen,
		a_rdata => wmi_rdata,
		a_en 		=> wmi_en
	);


	--
	-- Instance of SRAM-Access Component
	--

	Inst_sram_access: sram_access PORT MAP(
		clk 			=> clk,
		res 			=> res,

		-- I/O to SRAM (Onboard Hardware)
		sri_adr 	=> sri_addr,
		sri_data 	=> sri_data,
		sri_cs1 	=> sri_cs1,
		sri_cs2 	=> sri_cs2,
		sri_oe 		=> sri_oe,
		sri_ub 		=> sri_ub,
		sri_lb 		=> sri_lb,
		sri_we 		=> sri_we,

		-- Interface to SRAM Statemachine
		ri_adr 		=> sram_eff_addr,
		ri_din 		=> sr_din,
		ri_dout 	=> sr_dout,
		ri_rw 		=> sr_wen,
		ri_trg 		=> sr_trg,
		ri_rdy 		=> open
	);


---------------------------------------------
--
-- DMEM Address Generator
--
---------------------------------------------	

	process (clk,res, s_dm_adr, dmem_adr_offs)
	begin
		if res = '1' then
			dmem_adr_offs <= (others => '0');
			dm_adr			  <= (others => '0');

		elsif clk'event and clk = '1' then 
			-- DMEM Address Load
			if ena_set_dmem_offs = '1' then
				dmem_adr_offs <= dm_offs;
			end if;

			-- DMEM Address Increment
			if ena_dmem_offs_incr = '1' then
				dmem_adr_offs <= dmem_adr_offs + '1';
			end if;
		end if;

		-- Calculate DMEM Address (Counter + Base Address)
		dm_adr <= s_dm_adr + dmem_adr_offs;

	end process;


---------------------------------------------
--
-- SRAM Address Generator
--
---------------------------------------------	

	process (clk,res)
	begin
		if res = '1' then
			-- reset sram-addr-generator values
		  sram_base_addr 		<= (others => '0');
		  sram_base_count  	<= (others => '0');
		  sram_base_offset 	<= (others => '0');
		  sram_offs_mask   	<= (others => '0');

		elsif clk'event and clk = '1' then 

			-- enable load of sram-addr-generator values
			if load_sram_addr = '1' then
				-- load dm_rdata to specified register
			  case load_sram_dest is
				  when "00" 	=> sram_base_addr		<= dm_rdata (17 downto 0);
				  when "01" 	=> sram_base_count	<= dm_rdata (17 downto 0);
				  when "10" 	=> sram_base_offset <= dm_rdata (17 downto 0);
				  when "11" 	=> sram_offs_mask		<= dm_rdata (17 downto 0);
				  when others => NULL;
				end case;
			end if;

			-- clear offset register
			if clear_sram_offs = '1' then
			  sram_base_offset 	<= (others => '0');
			end if;

			-- increment counter
			if inc_sram_addr = '1' then
				sram_base_count <= sram_base_count + '1';
			end if;
		  
		end if;
	end process;

	-- calculate effective SRAM Address
	sram_eff_addr <= sram_base_addr + ((sram_base_count - sram_base_offset) and sram_offs_mask);

---------------------------------------------
--
-- Muxers
--
---------------------------------------------	
	
	-- input muxer
	process (clk,res)
	begin
		if res = '1' then
			in_muxer <= (others => '0');
		elsif clk'event and clk = '1' then 
			case in_src is
				when "00" => in_muxer <= input;
				when "01" => in_muxer <= signed (dm_rdata);
				when "10" => in_muxer <= out_muxer;
				when "11" => in_muxer <= accu;
				when others => NULL;
			end case;
			-- input also goes to sram
			sr_din <= std_logic_vector (in_muxer (23 downto 8));
		end if;
	end process;


	-- output muxer
	process (clk,res)
	begin
		if res = '1' then

		elsif clk'event and clk = '1' then 
			case out_src is
				when "00" => out_muxer <= accu;
				when "01" => out_muxer <= shape_out;
				when "10" => if sr_dout (15) = '1' then -- output from sram
												out_muxer <= signed (X"ff" & sr_dout & X"00");
										 else
												out_muxer <= signed (X"00" & sr_dout & X"00");
										 end if;
				when "11" => out_muxer <= input;
				when others => NULL;
			end case;
		end if;
	end process;



	-- DMEM Write Data muxer
	process (clk,res)
	begin
		if res = '1' then
			dmem_wdata <= (others => '0');
		elsif clk'event and clk = '1' then 
			case dmem_wdata_src is
				when "00" => dmem_wdata <= accu;
				when "01" => dmem_wdata <= shape_out;
				when "10" => dmem_wdata <= input;
				when "11" => dmem_wdata <= signed ("00" & X"000" & sram_base_count);
				when others => NULL;
			end case;
		end if;
	end process;

------------------------------------------------------------
--
--  Oscillator (Accumulator, Wave Register, Shaper, Random)
--
------------------------------------------------------------

	--
	-- Wave Register WREG
	--
	process (clk, res)
	begin
		if res = '1' then
		  wave_reg 		<= (others => '0');
		elsif clk'event and clk = '1' then
			if wave_reg_en = '1' then
			  wave_reg 		<= dm_rdata;
			end if;
		end if;
	end process;


  --
	-- Accumulator
	--
	process (clk,res)
	begin
		if res = '1' then
			accu <= (others => '0');
		elsif clk'event and clk = '1' then
      if ena_accu_load = '1' then
			  case accu_load_src is
				  when "00" 	=> accu <= (others => '0');
				  when "01" 	=> accu <= in_muxer;
					when others => accu <= (others => '0');
				end case;
			end if;

			if ena_accu = '1' then
				accu <= accu + in_muxer;
			end if;

		end if;
	end  process;

  
	
	--
	-- "Noise" Generator (may be improved)
	--
	process (clk,res)
	begin
		if res = '1' then
			noise_reg   <= (others => '0');
			noise_input <= '0';

		elsif clk'event and clk = '1' then
		  noise_input <= noise_reg (30) xnor noise_reg (16);

      if noise_shift_ena = '1' then
			  noise_reg <= noise_reg (30 downto 0) & noise_input;
			end if;

		end if;
	end  process;




	--
	-- Wave Shaper
	--
	process (clk,res)
	  variable accu8 : unsigned (7 downto 0);
	  variable pwm  : unsigned (7 downto 0);

	begin
		if res = '1' then
			shape_out <= (others => '0');
			wmem_adr 	<= (others => '0');
			accu8 := (others => '0');
			pwm   := (others => '0');
		elsif clk'event and clk = '1' then 
		  accu8 := conv_unsigned (in_muxer (31 downto 24), 8);
			pwm   := conv_unsigned (signed (wave_reg     (11 downto  4)), 8); 


			wmem_adr <= wave_reg (3 downto 2) & std_logic_vector (in_muxer (31 downto 24));


			if ena_shaper = '1' then

				case wave_reg (1 downto 0) is
				  -- sawtooth
				  when "00" 	=> 	shape_out <= 	in_muxer (31) & in_muxer (31) & in_muxer (31) & in_muxer (31) & 
																		 		in_muxer (31) & in_muxer (31) & in_muxer (31) & in_muxer (31) & in_muxer (31 downto 8);
					-- square (pwm)
				  when "01" 	=> 	if accu8 > pwm then 
														shape_out <= X"ff800000";
													else 
														shape_out <= X"007fffff";
												end if;
					-- wavetable
				  when "10" 	=> 	if wmem_data (15) = '1' then 
														shape_out <= signed (X"00" & wmem_data & X"00");
													else 
														shape_out <= signed (X"ff" & wmem_data & X"00");
													end if;
					-- noise
				  when "11" 	=> 	shape_out <= 	signed (noise_reg);
					when others => 	NULL;
				end case;
			end if;
		end if;
	end process;




---------------------------------------------
--
--  MicroCode "Decode"
--
---------------------------------------------	


	-- sram address generator signals

	load_sram_addr		 <= mm_rdata (28)							when enable_mc = '1' else '0';
	load_sram_dest		 <= mm_rdata (27 downto 26)		when enable_mc = '1' else (others => '0');
	clear_sram_offs		 <= mm_rdata (25)							when enable_mc = '1' else '0';
	inc_sram_addr			 <= mm_rdata (24)							when enable_mc = '1' else '0';

	-- dmem write register
	dmem_wdata_src			<= mm_rdata (23 downto 22)		when enable_mc = '1' else (others => '0');

	-- sram control
	sr_trg							<= mm_rdata (21)							when enable_mc = '1' else '0';
	sr_wen							<= mm_rdata (20)							when enable_mc = '1' else '0';

	-- dmem addr gen
	dm_offs							<= mm_rdata (19 downto 16)		when enable_mc = '1' else (others => '0');
	ena_set_dmem_offs		<= mm_rdata (15)							when enable_mc = '1' else '0';
	ena_dmem_offs_incr	<= mm_rdata (14)							when enable_mc = '1' else '0';

	-- dmem ctrl
	dm_wen							<= mm_rdata (13)							when enable_mc = '1' else '0';
	dm_en								<= mm_rdata (12)							when enable_mc = '1' else '0';

	-- inmux
	in_src 							<= mm_rdata (11 downto 10)		when enable_mc = '1' else (others => '0');

	-- outmux
	out_src							<= mm_rdata ( 9 downto  8)		when enable_mc = '1' else (others => '0');

	-- shaper
	ena_shaper					<= mm_rdata ( 7)							when enable_mc = '1' else '0';
	noise_shift_ena			<= mm_rdata ( 7)							when enable_mc = '1' else '0';  -- ena_shaper also triggers noise generator


	-- data register (load) enable
	wave_reg_en					<= mm_rdata ( 5)							when enable_mc = '1' else '0';

	-- accu
	ena_accu_load			 	<= mm_rdata ( 4)							when enable_mc = '1' else '0';
	ena_accu						<= mm_rdata ( 3)							when enable_mc = '1' else '0';
	accu_load_src				<= mm_rdata ( 2 downto  1)		when enable_mc = '1' else (others => '0');
											

	-- ctrl	
	ready_t							<= mm_rdata ( 0)							when enable_mc = '1' else '0';

	ready								<= ready_t										when enable_mc = '1' else '0';


---------------------------------------------
--
-- MicroCode Statemachine
--
---------------------------------------------	


	-- micro code
	process (clk,res)
	begin
	  if res = '1' then

			umc_state			<= (others => '0');
			mm_adr 				<= (others => '0');
			s_dm_adr 			<= (others => '0');
		  output 				<= (others => '0');

		elsif	clk'event and clk = '1' then

			case umc_state is
				-- wait for enable
				when X"0"  => 	if enable = '1' then
													mm_adr 					<= opcode & "00000";
													enable_mc				<= '0';	
													umc_state 			<= X"1";
													s_dm_adr 				<= am_adr;
--													s_cm_adr 				<= cm_adr;
												else
												  mm_adr 					<= '0' & X"00"; 
													umc_state 			<= X"0";
													enable_mc				<= '0';
												end if;

				when X"1"  =>  	umc_state 				<= X"2";
												enable_mc					<= '1';

				when X"2"  =>  	if ready_t = '1' then
												  mm_adr 					<= (others => '0');
													umc_state 			<= X"0";
													enable_mc				<= '0';
													output					<= out_muxer;
												else
													mm_adr <= mm_adr + '1';
													if mm_adr = '1' & X"FF" then
 													  umc_state	<= X"0";
														enable_mc			<= '0';
													end if;
												end if;

				when others => 	NULL;
			end case;

		end if;

	end process;


end Behavioral;


